package dev.fitko.fitconnect.client.bootstrap;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.config.defaults.Environments;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import static dev.fitko.fitconnect.api.config.defaults.Environments.PROD;
import static dev.fitko.fitconnect.api.config.defaults.Environments.STAGE;
import static dev.fitko.fitconnect.api.config.defaults.Environments.TEST;
import static dev.fitko.fitconnect.api.config.http.Timeouts.DEFAULT_TIMEOUT;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

class ApplicationConfigLoaderTest {

    @Test
    void testLoadMissingConfigFile() {
        Assertions.assertThrows(FitConnectInitialisationException.class, () -> ApplicationConfigLoader.loadConfigFromPath(Path.of("/no/path")));
    }

    @Test
    void testEnvironmentNotAvailable() throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-missing-env.yml");
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromYamlString(testConfigYaml);

        //When
        final FitConnectInitialisationException exception = Assertions.assertThrowsExactly(FitConnectInitialisationException.class, config::getCurrentEnvironment);

        //Then
        assertThat(exception.getMessage(), containsString("Environment 'ENVIRONMENT_THAT_DOES_NOT_EXIST' not found"));
    }

    @Test
    void testLoadConfigViaConfigBuilder() {

        // Given
        final var senderConfig = new SenderConfig("1", "123");

        final var subscriberConfig = SubscriberConfig.builder()
                .clientId("2")
                .clientSecret("123")
                .privateDecryptionKeyPaths(List.of())
                .privateSigningKeyPath("/dummy/path")
                .build();

        // When
        final ApplicationConfig config = ApplicationConfig.builder()
                .senderConfig(senderConfig)
                .subscriberConfig(subscriberConfig)
                .activeEnvironment(TEST.getEnvironmentName())
                .build();

        // Then
        assertNotNull(config);
        assertThat(config.getCurrentEnvironment(), equalTo(TEST.getEnvironment()));
        assertThat(config.getSenderConfig(), equalTo(senderConfig));
        assertThat(config.getSubscriberConfig(), equalTo(subscriberConfig));

        assertEquals(Environments.getEnvironmentsAsMap().size(), config.getEnvironments().size());
        assertThat(config.getEnvironments().get(PROD.getEnvironmentName()), is(PROD.getEnvironment()));
        assertThat(config.getEnvironments().get(STAGE.getEnvironmentName()), is(STAGE.getEnvironment()));
        assertThat(config.getEnvironments().get(TEST.getEnvironmentName()), is(TEST.getEnvironment()));
        assertThat(config.getActiveEnvironment(), is(TEST.getEnvironmentName()));

        assertFalse(config.getHttpConfig().getProxyConfig().isProxySet());
        assertThat(config.getHttpConfig().getTimeouts().getReadTimeoutInSeconds(), is(DEFAULT_TIMEOUT));
        assertThat(config.getHttpConfig().getTimeouts().getWriteTimeoutInSeconds(), is(DEFAULT_TIMEOUT));
        assertThat(config.getHttpConfig().getTimeouts().getConnectionTimeoutInSeconds(), is(DEFAULT_TIMEOUT));

        assertThat(config.getSubmissionDataSchemas().size(), is(0));
    }

    @Test
    void testLoadConfigViaConfigBuilderWithSubscriberJWKs() throws JOSEException {

        // Given

        final RSAKeyGenerator rsaKeyGenerator = new RSAKeyGenerator(2048);
        final JWK signingKey = rsaKeyGenerator.generate().toPublicJWK();
        final JWK decryptionKey = rsaKeyGenerator.generate().toPublicJWK();


        var subscriberConfig = SubscriberConfig.builder()
                .clientId("2")
                .clientSecret("123")
                .privateDecryptionKeys(List.of(decryptionKey))
                .privateSigningKey(signingKey)
                .build();

        // When
        final ApplicationConfig config = ApplicationConfig.builder()
                .subscriberConfig(subscriberConfig)
                .build();

        // Then
        assertNotNull(config);
        assertThat(config.getSubscriberConfig(), equalTo(subscriberConfig));
        assertThat(config.getSubscriberConfig().getSubscriberKeys().getPrivateSigningKey(), is(signingKey));
        assertThat(config.getSubscriberConfig().getSubscriberKeys().getPrivateDecryptionKeys(), is(List.of(decryptionKey)));
    }

    @Test
    void testLoadTestConfigFromString() throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");

        final Environment devEnv = new Environment(
                "https://auth-testing.fit-connect.fitko.dev",
                "https://routing-api-testing.fit-connect.fitko.dev",
                List.of("https://submission-api-testing.fit-connect.fitko.dev"),
                "https://portal.auth-testing.fit-connect.fitko.dev",
                true,
                false,
                false
        );

        final SenderConfig senderConfig = new SenderConfig("1", "123");

        // When
        final ApplicationConfig testConfig = ApplicationConfigLoader.loadConfigFromYamlString(testConfigYaml);

        // Then
        assertNotNull(testConfig);
        assertThat(testConfig.getCurrentEnvironment(), equalTo(devEnv));
        assertThat(testConfig.getSenderConfig(), is(senderConfig));
    }

    @Test
    void testLoadTestConfigFromPath(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");

        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig testConfig = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(testConfig);
        assertThat(testConfig.getActiveEnvironment(), is(new EnvironmentName("TEST")));
    }

    @Test
    void testSubscriberConfigContainsListOfPrivateDecryptionKeys(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig testConfig = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(testConfig);
        assertThat(testConfig.getSubscriberConfig().getPrivateDecryptionKeyPaths(), hasSize(2));
    }

    @Test
    void testDefaultConfigWithMinimalSettings(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-minimal-settings.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-with-minimal-settings.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then test defaults
        assertEquals(Environments.getEnvironmentsAsMap().size(), config.getEnvironments().size());
        assertThat(config.getEnvironments().get(PROD.getEnvironmentName()), is(PROD.getEnvironment()));
        assertThat(config.getEnvironments().get(STAGE.getEnvironmentName()), is(STAGE.getEnvironment()));
        assertThat(config.getEnvironments().get(TEST.getEnvironmentName()), is(TEST.getEnvironment()));
        assertThat(config.getActiveEnvironment(), is(TEST.getEnvironmentName()));

        assertFalse(config.getHttpConfig().getProxyConfig().isProxySet());
        assertThat(config.getHttpConfig().getTimeouts().getReadTimeoutInSeconds(), is(DEFAULT_TIMEOUT));
        assertThat(config.getHttpConfig().getTimeouts().getWriteTimeoutInSeconds(), is(DEFAULT_TIMEOUT));
        assertThat(config.getHttpConfig().getTimeouts().getConnectionTimeoutInSeconds(), is(DEFAULT_TIMEOUT));

        assertThat(config.getSubmissionDataSchemas().size(), is(0));
    }

    @Test
    void testUseDefaultEnvironmentWithHttpConfig(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-http-settings.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-with-http-settings.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertTrue(config.getHttpConfig().getProxyConfig().isProxySet());
        assertThat(config.getHttpConfig().getProxyConfig().getHost(), is("https://proxy.test.net"));
        assertThat(config.getHttpConfig().getProxyConfig().getPort(), is(8080));

        assertTrue(config.getHttpConfig().getProxyConfig().hasBasicAuthentication());
        assertThat(config.getHttpConfig().getProxyConfig().getBasicAuth().getUsername(), is("user"));
        assertThat(config.getHttpConfig().getProxyConfig().getBasicAuth().getPassword(), is("secret"));

        assertThat(config.getHttpConfig().getTimeouts().getReadTimeoutInSeconds(), is(Duration.ofSeconds(60)));
        assertThat(config.getHttpConfig().getTimeouts().getWriteTimeoutInSeconds(), is(Duration.ofSeconds(60)));
        assertThat(config.getHttpConfig().getTimeouts().getConnectionTimeoutInSeconds(), is(DEFAULT_TIMEOUT));
    }

    @Test
    void testOverridePropertiesOnDefaultEnvironment(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        assertTrue(TEST.getEnvironment().getAllowInsecurePublicKey());

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);
        final Environment envWithChangedDefaults = config.getEnvironments().get(TEST.getEnvironmentName());

        // Then
        assertFalse(envWithChangedDefaults.getAllowInsecurePublicKey());
    }

    @Test
    void testLoadingAdditionalCustomEnvironment(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-custom-env.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-with-custom-env.yml"), testConfigYaml);

        final Environment customEnvironment = new Environment();
        customEnvironment.setEnableAutoReject(true);
        customEnvironment.setAllowInsecurePublicKey(false);
        customEnvironment.setSkipSubmissionDataValidation(false);
        customEnvironment.setAuthBaseUrl("https://auth-url.test.net");
        customEnvironment.setRoutingBaseUrl("https://routing-url.test.net");
        customEnvironment.setSubmissionBaseUrls(List.of("https://submission-url-1.test.net", "https://submission-url-2.test.net"));
        customEnvironment.setSelfServicePortalBaseUrl("https://portal-url.test.net");

        // When
        final ApplicationConfig configWithCustomEnvironment = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(configWithCustomEnvironment);

        final Map<EnvironmentName, Environment> environments = configWithCustomEnvironment.getEnvironments();

        assertThat(environments.size(), is(Environments.getEnvironmentsAsMap().size() + 1));
        assertThat(environments.get(new EnvironmentName("CUSTOM_ENVIRONMENT")), is(customEnvironment));

        // currently there is only one url supported
        assertThat(configWithCustomEnvironment.getSubmissionBaseUrl(), is("https://submission-url-1.test.net"));

        assertThat(environments.get(PROD.getEnvironmentName()), is(PROD.getEnvironment()));
        assertThat(environments.get(STAGE.getEnvironmentName()), is(STAGE.getEnvironment()));
        assertThat(environments.get(TEST.getEnvironmentName()), is(TEST.getEnvironment()));
    }

    @Test
    void testLocalSubmissionDataSchemas(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(config.getSubmissionDataSchemas());
        assertThat(config.getSubmissionDataSchemas().size(), is(3));
    }

    @Test
    void testDefaultSchemaWriteVersions(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(config);
        assertThat(config.getMetadataSchemaWriteVersion(), is(URI.create("https://schema.fitko.de/fit-connect/metadata/1.2.0/metadata.schema.json")));
        assertThat(config.getSetSchemaWriteVersion(), is(URI.create("https://schema.fitko.de/fit-connect/set-payload/1.2.2/set-payload.schema.json")));
    }

    @Test
    void testAdjustDefaultSchemaWriteVersionsViaYaml(@TempDir final Path tempDir) throws IOException {

        // Given
        final String testConfigYaml = getResourceAsString("test-config-with-minimal-settings.yml");
        final Path configPath = Files.writeString(Path.of(tempDir.toString(), "test-config-with-minimal-settings.yml"), testConfigYaml);

        // When
        final ApplicationConfig config = ApplicationConfigLoader.loadConfigFromPath(configPath);

        // Then
        assertNotNull(config);
        assertThat(config.getMetadataSchemaWriteVersion(), is(URI.create("https://schema.fitko.de/fit-connect/metadata/1.2.0/metadata.schema.json")));
        assertThat(config.getSetSchemaWriteVersion(), is(URI.create("https://schema.fitko.de/fit-connect/set-payload/1.2.2/set-payload.schema.json")));
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        final File file = new File(Objects.requireNonNull(classLoader.getResource(name)).getFile());
        return Files.readString(file.toPath());
    }
}